import logging
from contextlib import asynccontextmanager
from dataclasses import dataclass, field, fields, replace
from inspect import signature
from typing import (Any, AsyncGenerator, AsyncIterator, Collection, Dict, Type,
                    TypeVar, Union, cast)

from .exception import ModelNotFoundException
from .fields import (deserialize, is_optional, is_transient, serialize,
                     update_field)
from .session import connection, transaction

_logger = logging.getLogger(__name__)


class ModelDataclassType(type):
    def __new__(cls, name, bases, dict):
        for field_name, field_type in dict.get("__annotations__", {}).items():
            update_field(field_name, field_type, dict)
        model_class = dataclass(
            super().__new__(cls, name, bases, dict), unsafe_hash=True
        )
        setattr(
            model_class,
            "NotFoundException",
            type("NotFoundException", (ModelNotFoundException,), {}),
        )
        return model_class


T = TypeVar("T", bound="Model")


class Model(metaclass=ModelDataclassType):
    id: int = field(init=True, repr=False, compare=False)

    @classmethod
    def prefix(cls) -> str:
        return f"{cls.__name__.lower()}"

    @classmethod
    async def get(cls: Type[T], id: Union[int, str]) -> T:
        async with connection() as conn:
            db_item = cast(Dict[str, Any], await conn.hgetall(f"{cls.prefix()}:{id}"))

        if not db_item:
            raise cls.NotFoundException(f"{id} not found")  # type: ignore # pylint: disable=no-member

        model_dict = {}
        for f in [f for f in fields(cls) if not is_transient(f)]:
            value = db_item.get(f.name)
            model_dict[f.name] = (
                None
                if is_optional(f) and value is None
                else await deserialize(f, value)
            )

        return cls(**model_dict)

    @classmethod
    def from_dict(cls: Type[T], model: Dict[str, Any], strict: bool = True) -> T:
        return (
            cls(**{k: v for k, v in model.items() if (k in signature(cls).parameters)})
            if strict
            else cls(**model)
        )

    @classmethod
    async def all(cls: Type[T]) -> AsyncIterator[T]:
        async with connection() as conn:
            for key in await conn.smembers(cls.prefix()):
                value = cast(T, await cls.get(key))
                if value:
                    yield value
                else:
                    _logger.warning(f"{cls.__name__} Key: {key} orphaned")

    @staticmethod
    async def flush():
        async with connection() as conn:
            await conn.flushdb()

    @classmethod
    async def count(cls) -> int:
        async with connection() as conn:
            return await conn.scard(cls.prefix())

    @classmethod
    async def delete_all(cls: Type):
        key_prefix = cls.prefix()
        async with connection() as conn:
            keys = await conn.keys(f"{key_prefix}:*")
            await conn.delete(key_prefix, *keys)

    @classmethod
    async def persisted(cls: Type, id: int) -> bool:
        async with connection() as conn:
            return await conn.exists(f"{cls.prefix()}:{id}")

    @property
    def db_id(self) -> str:
        return f"{self.prefix()}:{self.id}"

    async def save(self, optimistic=False):
        async with self._serialized_model(optimistic) as model_dict:
            async with transaction() as tr:
                tr.hmset_dict(self.db_id, model_dict)
                tr.sadd(self.prefix(), self.id)

    async def update(self, optimistic=False, **changes):
        async with self._serialized_model(optimistic, **changes) as model_dict:
            async with transaction() as tr:
                for key, value in model_dict.items():
                    tr.hset(self.db_id, key, value)
                return replace(self, **changes)

    @asynccontextmanager
    async def _serialized_model(
        self, optimistic, **changes
    ) -> AsyncGenerator[Dict[str, Any], None]:
        async with connection() as conn:
            if optimistic:
                conn.watch(self.db_id)
            model_dict = {}
            for field in [
                f
                for f in fields(self)
                if not is_transient(f)
                and hasattr(self, f.name)
                and (not changes or f.name in changes)
            ]:
                value = changes.get(field.name, getattr(self, field.name))
                ref_key = f"{self.db_id}:{field.name}"
                serialized = await serialize(field, ref_key, value)
                if isinstance(serialized, (Collection, Model)) and not isinstance(
                    serialized, str
                ):
                    model_dict[field.name] = ref_key
                    await serialized.save(optimistic=optimistic)
                elif serialized:
                    model_dict[field.name] = serialized
            yield model_dict

    async def delete(self):
        key = self.db_id
        async with connection() as conn:
            keys = await conn.keys(f"{key}:*")
            await conn.delete(*keys, key)

    async def exists(self) -> bool:
        async with connection() as conn:
            return await conn.exists(self.db_id)

    async def refresh(self: T) -> T:
        refreshed = await type(self).get(self.id)
        return cast(T, refreshed)



from contextlib import asynccontextmanager
from contextvars import ContextVar

from aioredis.commands import ContextRedis, Redis, create_redis_pool
from aioredis.commands.transaction import MultiExec

REDIS: ContextVar[Redis] = ContextVar("redis", default=None)
CONNECTION: ContextVar[ContextRedis] = ContextVar("connection", default=None)
TRANSACTION: ContextVar[MultiExec] = ContextVar("transaction", default=None)

config = {}


def configure(address: str = None, *args, **kwargs):
    global config
    config["address"] = address
    config["args"] = args
    config["kwargs"] = kwargs


@asynccontextmanager
async def redis_pool(address: str = None, *args, **kwargs):
    redis = REDIS.get()
    if redis:
        yield redis
        return
    redis = await create_redis_pool(
        address if address else config.get("address", "redis://localhost"),
        encoding=kwargs.get("encoding", "utf-8"),
        *args if args else config.get("args", []),
        **kwargs if kwargs else config.get("kwargs", {})
    )
    t = REDIS.set(redis)
    try:
        yield redis
    finally:
        redis.close()
        await redis.wait_closed()
        REDIS.reset(t)


@asynccontextmanager
async def connection(address: str = None, *args, **kwargs):
    connection = CONNECTION.get()
    if connection:
        yield connection
        return

    async with redis_pool(address, *args, **kwargs) as redis:
        with await redis as conn:
            t = CONNECTION.set(conn)
            yield conn
            CONNECTION.reset(t)


@asynccontextmanager
async def transaction():
    transaction = TRANSACTION.get()
    if transaction:
        yield transaction
        return

    async with connection() as conn:
        transaction = conn.multi_exec()
        t = TRANSACTION.set(transaction)
        try:
            yield transaction
        finally:
            await transaction.execute()
            TRANSACTION.reset(t)



import json
from abc import ABCMeta, abstractmethod
from asyncio.coroutines import iscoroutine
from typing import (TYPE_CHECKING, AbstractSet, Any, Collection, Dict, Generic,
                    Iterator, MutableSequence, MutableSet, Optional, Sequence,
                    Type, TypeVar, Union, cast)

if TYPE_CHECKING:
    from .model import Model

from .exception import ModelNotLoadedException
from .session import connection, transaction

T = TypeVar("T", bound="Model")


class RedisCollection(Collection, metaclass=ABCMeta):
    def __init__(self, key: str, values: Sequence):
        self._key = key
        self.values = values

    @classmethod
    async def from_key(cls, key: str, eager: bool = False, **kwargs):
        values = await cls.get_values_for_key(key) or []
        collection = cls(
            key, type(values)([json.loads(value) for value in values]), **kwargs
        )
        if eager:
            async for _ in collection:
                pass
            return collection.values
        return collection

    @abstractmethod
    async def save(self, optimistic: bool = False):
        pass

    @classmethod
    @abstractmethod
    async def get_values_for_key(cls, key: str) -> Sequence:
        pass

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, value):
        self._values = value

    def __iter__(self) -> Iterator:
        return filter(None, iter(self.values))

    def __len__(self) -> int:
        return len(self.values)

    def __contains__(self, x) -> bool:
        return x in self.values

    def __aiter__(self):
        self._iter = iter(self.values)
        return self

    async def __anext__(self):
        try:
            val = next(self._iter)
            if iscoroutine(val):
                return await val
            return val
        except StopIteration:
            raise StopAsyncIteration


class RedisSet(MutableSet, RedisCollection):
    @classmethod
    async def get_values_for_key(cls, id: int) -> AbstractSet:
        async with connection() as conn:
            return set(await conn.smembers(id))

    async def save(self, optimistic: bool = False):
        async with connection() as conn:
            if optimistic:
                conn.watch(self._key)
            async with transaction() as tr:
                tr.delete(self._key)
                tr.sadd(self._key, *super().values)

    def add(self, value):
        self.values.append(value)

    def discard(self, value):
        self.values.discard(value)


class RedisList(MutableSequence, RedisCollection):
    @classmethod
    async def get_values_for_key(cls, id: int) -> Sequence:
        async with connection() as redis:
            return await redis.lrange(id, 0, -1)

    async def save(self, optimistic: bool = False):
        async with connection() as conn:
            if optimistic:
                conn.watch(self._key)
            async with transaction() as tr:
                tr.delete(self._key)
                tr.rpush(self._key, *super().values)

    def insert(self, index: int, value: T):
        self.values.insert(index, value)

    def __getitem__(self, i: Union[int, slice]) -> Union[Any, Sequence[Any]]:
        return self.values.__getitem__(i)

    def __setitem__(self, i: Union[int, slice], o: Union[Any, Sequence[Any]]):
        self.values.__setitem__(i, o)

    def __delitem__(self, i: int):
        self.values.__delitem__(i)


class ModelCollection(RedisCollection, Generic[T], metaclass=ABCMeta):
    def __init__(
        self, key: str, keys: Sequence[int], model_class: Type[T] = None, cascade=False
    ):
        self._cache: Dict[int, T] = {}
        self._collection_type = type(keys)
        self._model_class = model_class
        self._cascade = cascade
        super().__init__(key, keys)

    async def save(self, optimistic: bool = False):
        async with transaction():
            await super().save(optimistic=optimistic)
            if self._cascade:
                async for v in self:
                    await v.save(optimistic=optimistic)

    @property
    def values(self):
        return self._collection_type(self._cache.values())

    @values.setter
    def values(self, values):
        if values and isinstance(next(iter(values)), int):  # Lazy loading using keys
            super(ModelCollection, type(self)).values.fset(
                self, self._collection_type(values)
            )
        else:  # values: Collection[Model]
            super(ModelCollection, type(self)).values.fset(
                self, self._collection_type([v.id for v in values])
            )
            self._cache = {v.id: v for v in values}

    async def _get_item(self, key: int) -> Optional[T]:
        if key in self._cache:
            return self._cache[key]
        value = cast(T, await self._model_class.get(key))
        self._cache[key] = value
        return value

    def __aiter__(self):
        self._iter = map(self._get_item, super().values)
        return self


class RedisModelSet(ModelCollection[T], RedisSet):
    def add(self, value: T) -> None:
        key = value.id
        super().add(key)
        self._cache[key] = value

    def discard(self, value: T) -> None:
        key = value.id
        super().discard(key)
        self._cache.pop(key)


class RedisModelList(ModelCollection[T], RedisList):
    def _get_cached_item(self, key: int):
        try:
            return self._cache[key]
        except KeyError:
            raise ModelNotLoadedException(
                f"{self._model_class.__name__} object with id {key} was not loaded"
            )

    def insert(self, index: int, value: T) -> None:
        key = value.id
        super().insert(index, key)
        self._cache[key] = value

    def __getitem__(self, i: Union[int, slice]) -> Union[T, Sequence[T]]:
        if isinstance(i, slice):
            return [
                self._get_cached_item(self.values[index])
                for index in range(*i.indices(len(self)))
            ]
        else:
            return self._get_cached_item(self.values[i])

    def __setitem__(self, i: Union[int, slice], o: Union[T, Sequence[T]]) -> None:
        if isinstance(i, slice) and isinstance(o, Sequence):
            for obj_index, index in enumerate(range(*i.indices(len(self)))):
                self.insert(index, o[obj_index])
        elif not isinstance(i, slice) and not isinstance(o, Sequence):
            self.insert(i, o)
        else:
            raise ValueError

    def __delitem__(self, i: int) -> None:
        key = self.values[i]
        super().__delitem__(i)
        del self._cache[key]



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="aio-rom",
    version="0.0.5",
    author="Federico Jaite",
    author_email="fede_654_87@hotmail.com",
    description="asyncio based Redis object mapper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fedej/aio-rom",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Topic :: Database :: Front-Ends",
    ],
    python_requires='>=3.7',
    install_requires=[
        "aioredis>=1.3.1",
        "typing-inspect>=0.6.0"
    ],
)
